#ifndef VECTOR3
#define VECTOR3

#include <cmath>
#include <ostream>

//!Class representing 3D vectors, with elements of type T
template<class T>
class Vec3 {
public:
	T x;
	T y;
	T z;

public:
	//!Constructor accepting a single value, which defaults to 0. All three components are set to the same value.
	Vec3(T value=0) : x(value), y(value), z(value) {}

	//!Constructor accepting three values, one for each vector component.
	Vec3(T _x, T _y, T _z) : x(_x), y(_y), z(_z) {}

	//!Copy constructor
	template<class T2>
	Vec3(Vec3<T2> v): x(v.x), y(v.y), z(v.z) {}

	//!Vector addition
    Vec3<T> operator+(const Vec3<T> &v) const {
      return Vec3<T>(x+v.x,y+v.y,z+v.z);
    }

	//!In-place vector addition
    Vec3<T>& operator+=(const Vec3<T> &v) {
      x += v.x;
      y += v.y;
      z += v.z;
      return (*this);
    }

	//!Vector subtraction
    Vec3<T> operator-(const Vec3<T> &v) const {
      return Vec3<T>(x-v.x,y-v.y,z-v.z);
    }

	//!In-place vector subtraction
    Vec3<T>& operator-=(const Vec3<T> &v) {
      x -= v.x;
      y -= v.y;
      z -= v.z;
      return (*this);
    }

	//!Multiplication with double
	Vec3<T> operator*(const double &d) const {
      return Vec3<T>(T(x*d),T(y*d),T(z*d));
    }

	//!In-place multiplication with double
    Vec3<T>& operator*=(const double &d) {
      x = T(x*d);
      y = T(y*d);
      z = T(z*d);
      return (*this);
    }

	//!Division with double
    Vec3<T> operator/(const double &d) const {
      return Vec3<T>(T(x/d),T(y/d),T(z/d));
    }
	
	//!In-place division with double
	Vec3<T>& operator/=(const double &d) {
      x = T(x/d);
      y = T(y/d);
      z = T(z/d);
      return (*this);
    }

	//!Scalar product
    double operator*(const Vec3<T> &v) const {
      return x*v.x + y*v.y + z*v.z;
    }

	//!Cross product
    Vec3<T> cross(const Vec3<T> &v) const {
      Vec3<T> tmp;
      tmp.x=y*v.z-z*v.y;
      tmp.y=z*v.x-x*v.z;
      tmp.z=x*v.y-y*v.x;
      return tmp;
    }

	//!Element-wise multiplication (hadamard product)
	 Vec3<T> operator%(const Vec3<T> &v) const {
      return Vec3<T>(x*v.x, y*v.y, z*v.z);
    }

	//!Normalize
    Vec3<T>& normalize() {
      double n = norm();
      (*this) /= n;
      return (*this);
    }

	//!Vector norm
    double norm() const {
      return std::sqrt(double(x*x+y*y+z*z));
    }
    
    //!Vector norm squared
    double norm2() const {
      return x*x+y*y+z*z;
    }

	//!Equal (used for integer position vectors)
    bool operator==(const Vec3<T> &v) const {
      return (x==v.x && y==v.y && z==v.z);
    }

    //!Not equal (used for integer position vectors)
    bool operator!=(const Vec3<T> &v) const {
      return !((*this)==v);
    }

	 //!Multiplication with double
    friend Vec3<T> operator*(const double &d, const Vec3<T> &v) {
      return v*d;
    }

	//!Cast to scalar, simply returns the x component
	operator T() { return x; }

	//!Write to ostream
	friend std::ostream& operator<<(std::ostream &os, const Vec3<T> &v) {
      os << v.x << " " << v.y<<" "<<v.z;
      return os;
    }



};



template class Vec3<float>;
template class Vec3<double>;
template class Vec3<int>;

typedef Vec3<float>     Vec3f;
typedef Vec3<double>    Vec3d;
typedef Vec3<int>       Vec3i;

#endif
